%--------------------------------------------------------------------------
% Bounded Distortion Mesh Parameterization
%
% Run demo.m this to see an example of how to flatten a disk-type mesh while minimizing 
% the maximal conformal distortion using the BD spaces in:
% "Bounded Distortion Mapping Spaces For Triangular Meshes" 
% Y. Lipman.
% and the method in
% "Controlling Singular Values with Semidefinite Programming"
% Shahar Kovalsky, Noam Aigerman, Ronen Basri, Yaron Lipman
% 
% This piece of code, written by N. Aigerman, S. Kovalsky, and Y. Lipman
%
% Disclaimer: The code is provided as-is and without any guarantees. 
%--------------------------------------------------------------------------



%% init
rng(1)
close all
clear
initialize;
yalmip('clear')

%% parameters
prm.debug = true;
prm.visualize = true;

prm.init.method = 'solve_unconstrained';
prm.constraints.C = 1000; 
prm.solve.solver = 'mosek'; %solver of choice
prm.solve.max_iter = 1000;
prm.solve.thresh_obj = 1e-6;
prm.solve.thresh_Y = 1e-6;
prm.solve.yalmip_prm = sdpsettings('solver',prm.solve.solver,'verbose',0);
prm.solve.yalmip_prm.quadprog.Algorithm = 'interior-point-convex';

% prm = setParams;
MAX_ITER = 150;

%% load mesh

mesh_path = '.\models\';

% meshname='foot_decim.obj';
meshname = 'cathead.obj';
%  meshname='fandisk_decim.obj'; 
%  meshname='tooth.obj'; 
%  meshname='hand.obj';
%  meshname='isis_decim.obj';
%  meshname='bunny_decim.obj';
%  meshname='dino_decim.obj';

OBJ = read_wobj([mesh_path meshname]);
X = OBJ.vertices;
if(length(OBJ.objects)==4)
    tri = OBJ.objects(4).data.vertices;%for meshlab objs
else
    tri = OBJ.objects.data.vertices;%for standard objs
end
        
TR = triangulation(tri,X);

FBtri = freeBoundary(TR);
anchor_ind = FBtri(:,1);
tet = linspace(0,2*pi-eps,length(anchor_ind)+1);
anchor_coords = [cos(tet') sin(tet')];
anchor_coords(end,:)=[];


%% initialize variables
[Y, Faces] = mesh3DToYalmip(X,tri,prm);


%% solve Tutte embedding for initial guess (assume nice angles of source mesh)

n = size(X,1);
ii = [tri(:,1) ; tri(:,2); tri(:,3)];
jj = [tri(:,2) ; tri(:,3); tri(:,1)];
ss = ones(size(ii));
L = sparse(ii,jj,ss,n,n);
L = double(L|L');
L = double(diag(sum(L,2)) -  L);
L(anchor_ind,:)=0;
L = L + sparse(anchor_ind,anchor_ind,ones(length(anchor_ind),1),n,n);
b = zeros(n,2);
b(anchor_ind,:) = anchor_coords;
Y_hat = L\b;
assign(Y,Y_hat);

%update frames
[Faces,distortions,flips] = updateFrames(Faces);
Y_hat = double(Y);

%% positional constraints 
%for fixed boundary uncomment:
% pos_constraints = Y(anchor_ind,:)==anchor_coords; %fix all boundary verts

%for free boundary uncomment: (fix a point and a direction) 
tempz = [1 -1].'; tempz = [real(tempz) imag(tempz)];
[~,ind1] = min( sum( (Y_hat(anchor_ind,:) -  ones(size(anchor_ind))*tempz(1,:)).^2 ,2) ); ind1=ind1(1);
[~,ind2] = min( sum( (Y_hat(anchor_ind,:) -  ones(size(anchor_ind))*tempz(2,:)).^2 ,2) ); ind2=ind2(1);
temp = Y_hat(anchor_ind(ind2),:)-Y_hat(anchor_ind(ind1),:); 
n_hat = [-temp(2); temp(1)];
pos_constraints =  (Y(anchor_ind(ind1),:) == anchor_coords(ind1,:)) + ...
    (  (Y(anchor_ind(ind2),:)-Y(anchor_ind(ind1),:))*n_hat == 0   );




%% visualize initiailzation
figure(1);
visualizeMesh(tri, Y_hat);
axis equal;

cax = [1 4];%color axis (corresponds to distortion)
edgecolor = 'none'; %edge color: do not plot edges

%draw the UV, and save UV-mesh to VTK file (can be shown in Paraview software)
QCONF_display_param_result(X,tri,Y_hat,cax,edgecolor,'Tutte',101,'mesh_uv_tutte');



%% minimize L_inf conformal distortion
m = length(Faces);
dl = sdpvar(1,m);% a new variable per face
t = sdpvar(1,1);% a global variable, used for max optimization

t_hat = -inf;
iter_num = 0;

e = sdpvar(1,1);


while abs(t_hat) > 10^-5 && iter_num < MAX_ITER

    
    %build the distortion constraints
    temp = [Faces{:}];
    A = reshape([temp.FA],2,2,[]);
    a = squeeze(A(1,1,:))';
    b = squeeze(A(1,2,:))';
    c = squeeze(A(2,1,:))';
    d = squeeze(A(2,2,:))';%the entries of A
    
    alpha = [a+d;b-c]/2; %2XM
    beta = [a-d;b+c]/2;
    
    % the cone |beta| <= mu*Re(alpha) 
    C = distortions';
    maxC = max(C);
    
    %for Linf 
    cone_BD = cone( [( ( (maxC-1)./(maxC+1) ) .* alpha(1,:) + t ) ; beta ] );
           
    constraints  = cone_BD;
    
    
    %build the objective
    objective = t;
    

    %optimize
    res = solvesdp(pos_constraints+constraints, objective, prm.solve.yalmip_prm);
    
    
    distortions_old = max(distortions);
    [Faces,distortions,flips] = updateFrames(Faces);
    %output
    Y_old = Y_hat;
    Y_hat = double(Y);
    t_hat = abs(max(distortions) - distortions_old); 
%     t_hat = norm(Y_hat - Y_old,'fro')/length(Y_hat);
    iter_num = iter_num+1;
    
    str = sprintf('iter_num: %d t_hat %f', iter_num, t_hat);
    disp(str);
    
    disp(['maximal conformal distortion: ' num2str(max(distortions))]);
    disp(['mean conformal distortion: ' num2str(mean(distortions))]);
    disp(['median conformal distortion: ' num2str(median(distortions))]);
    
end

%% visualize
figure(2);
visualizeMesh(tri, Y_hat);
axis equal;

% % TEETH_output_vtk_texture('mytemp.vtk',X,tri,Y_hat)
cax = [1 4];%color axis (corresponds to distortion)
edgecolor = 'none'; %edge color: do not plot edges

%draw the UV, and save UV-mesh to VTK file (can be shown in Paraview software)
QCONF_display_param_result(X,tri,Y_hat,cax,edgecolor,'Bounded-Conformal Distortion',103,'mesh_uv_results');


